package SM
is
   --  SPARK is pedantic about overflows, so we will
   --  constrain both the length of the input array,
   --  and the range of the element values, so we can
   --  construct a proof for the "no overflow" property.

   --  Rather arbitrary choices here, but chosen so that
   --  Max_N * Max_Value <= Integer'Last
   Max_N     : constant := 100;
   Max_Value : constant := 1000;

   subtype Index is Natural range 1 .. Max_N;
   subtype Value is Natural range 0 .. Max_Value;

   --  Note this type need to include 0 since that is
   --  the identity value for arithmetic "+"
   subtype Sum_Range is Natural range 0 .. (Max_N * Max_Value);

   type Vector is array (Index range <>) of Value;

   --==========================================
   --  Declaration of abstract proof functions
   --==========================================

   --  Sum_Of (A, L, U) returns the arithmetic sum of
   --  the elements of A between index values L and U

   --# function Sum_Of (A          : in Vector;
   --#                  From_Bound : in Index;
   --#                  To_Bound   : in Index) return Sum_Range;

   --  Max_Of (A, L, U) returns the maximum value of the
   --  elements of A between index values L and U

   --# function Max_Of (A          : in Vector;
   --#                  From_Bound : in Index;
   --#                  To_Bound   : in Index) return Value;

   --========
   --  AXIOMS
   --========

   -- These are written in FDL - the proof logic used by the SPARK
   -- proof tools.  These are actually read into the Simplifier
   -- in the file sm/sm.rlu and for the Checker in sm/sm.rul

   -- 1 and 2 - sum and max of unit range
   -- sm_rules(1): sum_of(A, F, F) may_be_replaced_by element(A, [F]).
   -- sm_rules(2): max_of(A, F, F) may_be_replaced_by element(A, [F]).

   -- 3 and 4 - sum and max of the null range
   -- sm_rules(3): sum_of(A, F, L) may_be_replaced_by 0 if [L < F].
   -- sm_rules(4): max_of(A, F, L) may_be_replaced_by 0 if [L < F].

   -- 5 - definition of Sum_Of for I+1 elements in terms of Sum_Of I elements
   -- sm_rules(5): sum_of(A, F, I+1) may_be_replaced_by sum_of(A, F, I) + element(A, [I+1]).

   -- 6 - properties of Max_Of the I+1'th element - case 1 - I+1'th element is less
   --     than or equal to the Max_Of the first I elements
   -- sm_rules(6): max_of(A, F, I+1) may_be_replaced_by max_of(A, F, I) if
   --           [element(A, [I+1]) <= max_of(A, F, I)].

   -- 7 - properties of Max_Of the I+1'th element - case 2 - I+1'th element is greather
   --     than the Max_Of the first I elements
   -- sm_rules(7): max_of(A, F, I+1) may_be_replaced_by element(A, [I+1]) if
   --           [element(A, [I+1]) > max_of(A, F, I)].

   --================================
   --  Exported procedures
   --================================

   --  Unconstrained array type here (like a "conformant array
   --  parameter in Level 1 ISO Pascal (for older readers)).  The actual
   --  size of the actual parameter might be different for each call
   --  and the actual bounds are passed as additional attributes
   --  A'First and A'Last of the formal parameter.
   procedure Sum_Max (A   : in     Vector;
                      Sum :    out Sum_Range;
                      Max :    out Value);
   --# derives Sum, Max from A;
   --# post Sum = Sum_Of (A, A'First, A'Last) and
   --#      Max = Max_Of (A, A'First, A'Last);

end SM;



