(* libraries etc *)
val _ = app load ["finite_mapTheory","wordsTheory","unwindLib","stringTheory"]; 

use "tactic.sml"; open tactic; (* EXPORT tactic.sml *)

val failwith_def = Define `
failwith (s:string) = (ARB:'a)
`;

(******************************************************************************)
(* defns *)

(* main development *)

val _ = type_abbrev("q_ty", ``:'a list # 'a list``);

val front_def = Define `front = FST`;

val rear_def = Define `rear = SND`;

val mk_q_def = Define `
mk_q f r = (f,r)
`;

val dest_q_def = Define `
dest_q (f,r) = (f,r)
`;

(* unique repn. of empty q *)
val empty_def = Define `
empty = mk_q [] []
`;

(* the invariant for the q subtype *)
val qinv_def = Define `
qinv q = LENGTH (rear q) <= LENGTH (front q)
`; 

val collapse_def = Define `
collapse q = 
  let (f,r) = dest_q q in
  mk_q (f++REVERSE r) []
`;

val Enqueue_def = Define `
Enqueue x q = 
  if q = empty then mk_q [x] [] else
  let (f,r) = 
    if LENGTH (rear q) = LENGTH (front q) then 
      dest_q (collapse q)
    else 
      dest_q q
  in
  (* need to maintain invariant *)
  mk_q f (x::r)
`;

val Front_def = Define `
Front q = 
  if front q = [] then 
    (* invariant implies q = empty *) 
    failwith "Front: q is empty" 
  else HD (front q)
`;

val Tail_def = Define `
Tail q = 
  if q = empty then failwith "Tail: q is empty" else
  let (f,r) = 
    if LENGTH (rear q) = LENGTH (front q) then
      (* need to maintain invariant, can't just take tail of f *)
      dest_q (collapse q)
    else
      dest_q q
  in
  mk_q (TL f) r
`;


(* abstract to a simple list *)
val abstr_def = Define `
abstr q = (front q) ++ (REVERSE (rear q))
`;


(******************************************************************************)
(* proofs *)

(* comment on the proof format

proofs look like this:

xxx;
yyy;
zzz;

--

solved goals are followed by a blank line eg

xxx;
solve_goal_tactic;

yyy; (* start on next proof *)
zzz;

--

if you increase the number of goals, you indent by one space eg

xxx;
create_two_subgoals;
 solve_subgoal_1;

 solve_subgoal_2;


block structure is preserved, ie subgoals of a given goal all have the
same indentation level

--

explicit facts are stated using "have" eg

have `some facts`;

--

tactics which solve goals outright are usually "tac" or "ssimp[...]
THEN NO_TAC"

--

main tactics are ssimp (simplification) and tac (first order proof)

NO_TAC is the tactic that always fails (if there are subgoals!)

Cases_on does a case analysis (typically on a list, or a proposition)

*)


val defs1 = [front_def,rear_def,mk_q_def,dest_q_def,empty_def,qinv_def];
val defs2 = [collapse_def,Enqueue_def,Front_def,Tail_def];
val defs3 = [abstr_def];

(* library stuff *)

g `
! x y. 
(x <= SUC y /\ x <> SUC y ==> x <= y)
/\ (x <= y /\ x <> y ==> SUC x <= y)
/\ (1 <= SUC x + SUC y)
`;
e(numLib.ARITH_TAC);
val le_thm = top_thm();


(* simple stuff *)
g `
(q = empty) = (abstr q = [])
`;
e(Cases_on `q`);
e(ssimp (defs1@defs3) THEN NO_TAC);
val q_empty_abstr = top_thm();

g `
qinv (f,r) ==> (f = []) ==> (r = [])
`;
ir();
ir();
e(ssimp defs1);
e(Cases_on `r` THEN ssimp[] THEN NO_TAC);
val f_nil_r_nil = top_thm();

(* qinv stuff *)

g `
qinv q ==> qinv (Enqueue x q)
`;
ir();
e(Cases_on `q`);
e(RENAME_TAC [``q':'a list``|->``f:'a list``]);
e(Cases_on `f=[]`); 
 (* f = [] *)
 have `r=[]`; e(tac[f_nil_r_nil]);
 e(ssimp (defs1@defs2) THEN NO_TAC);

 (* f <> [] *)
 e(Cases_on `LENGTH r = LENGTH f`);
  e(ssimp (defs1@defs2@[LET_THM]));
  e(Cases_on `r` THEN Cases_on `f` THEN ssimp[]);
  e(tac[le_thm]);

  e(ssimp (defs1@defs2@[LET_THM]));
  e(tac[le_thm]);
val qinv_Enqueue = top_thm();

(* Front doesn't affect the list *)

g `
qinv q ==> q <> empty ==> qinv (Tail q)
`;
ir();
ir();
e(Cases_on `q`);
e(RENAME_TAC [``q':'a list``|->``f:'a list``]);
have `f <> []`;
 e(ssimp defs1);
 e(Cases_on `f` THEN Cases_on `r` THEN ssimp[] THEN NO_TAC);
e(Cases_on `LENGTH r = LENGTH f`);
 e(ssimp (defs1@defs2@[LET_THM]) THEN NO_TAC);

 e(ssimp (defs1@defs2@[LET_THM]));
 e(Cases_on `f` THEN ssimp[]);
 e(tac[le_thm]);
val qinv_Tail = top_thm();



(* abstr proofs *)
g `
qinv q ==> (abstr (Enqueue x q) = ((abstr q) ++ [x]))
`;
ir();
e(Cases_on `q`);
e(RENAME_TAC [``q':'a list``|->``f:'a list``]);
e(Cases_on `f=[]`);
 (* f = [] *)
 have `r=[]`; e(tac[f_nil_r_nil]);
 e(ssimp (defs1@defs2@defs3@[LET_THM]) THEN NO_TAC);

 (* f <> [] *)
 e(Cases_on `LENGTH r = LENGTH f`);
  e(ssimp (defs1@defs2@defs3@[LET_THM]) THEN NO_TAC);

  e(ssimp (defs1@defs2@defs3@[LET_THM]) THEN NO_TAC);
val abstr_Enqueue = top_thm();

g `
qinv q ==> (abstr q = x::xs) ==> (Front q = x)
`;
ir();
ir();
e(Cases_on `q`);
e(RENAME_TAC [``q':'a list``|->``f:'a list``]);
have `f <> []`;
 ir();
 e(ssimp (defs1@defs2@defs3@[LET_THM]));
 have `r=[]`; e(Cases_on `r` THEN ssimp[] THEN NO_TAC);
 e(ssimp[] THEN NO_TAC);
e(Cases_on `f`);
 e(ssimp[] THEN NO_TAC);
e(ssimp (defs1@defs2@defs3@[LET_THM]) THEN NO_TAC);
val abstr_Front = top_thm();

g `
qinv q ==> (abstr q = x::xs) ==> (abstr (Tail q) = xs)
`;
ir();
ir();
e(Cases_on `q`);
e(RENAME_TAC [``q':'a list``|->``f:'a list``]);
have `f <> []`;
 ir();
 e(ssimp (defs1@defs2@defs3@[LET_THM]));
 have `r=[]`; e(Cases_on `r` THEN ssimp[] THEN NO_TAC);
 e(ssimp[] THEN NO_TAC);
have `(f,r) <> empty`;
 e(ssimp defs1 THEN NO_TAC);
e(Cases_on `LENGTH r = LENGTH f`);
 e(ssimp (defs1@defs2@defs3@[LET_THM]) THEN NO_TAC);

 e(ssimp (defs1@defs2@defs3@[LET_THM]));
 e(Cases_on `f` THEN ssimp[] THEN NO_TAC);
val abstr_Tail = top_thm();
